######################################################################
#:AssocList
#
# Licensed Materials - Property of IBM
# Rational ClearCase
# (C) Copyright IBM Corp. 2001, 2008 All Rights Reserved
# US Government Users Restricted Rights
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp
#
# Copyright 2001 Rational Software Corporation. All Rights Reserved
# This software contains proprietary and confidential information of
# Rational and its suppliers.  Use, disclosure or reproduction is
# prohibited without the prior express written consent of Rational.

=head1 NAME

CQCC::AssocList - Class to manage a list of ClearQuest Associations 

=head1 CLASS DESCRIPTION

This class manages a list of ClearQuest Change Requests that can be
associated with a ClearCase version. The entity, database, and number
are stored for each item, providing initial support for multiple
entities and multiple databases within a list. Methods provide the
same information formatted different ways for different needs; handle
parsing user input and output; and compare two lists to identify the
differences. 

SUPPORT POLICY: See TriggerCQCC.pm for the Rational Support Policy. 

=head1 METHODS

=cut

######################################################################
#::Package CQCC::AssocList
######################################################################

package CQCC::AssocList;
use strict;

use CQCC::IOBase;     # Include in case we need to make own copy
use CQCC::Parm;       # Implicitly included in iobase

sub Version { return '2.1' };

######################################################################
#::Class Globals 
######################################################################

######################################################################
#::function ClassInit
# Called implicitly when new object is created
######################################################################

# sub ClassInit {}
# &ClassInit();   

######################################################################
#::function ClassTest

=over 6

=item C<ClassTest(args)>

This function is a local unit test that will create an instance of the
Class and ask it to parse a typical user input string to set the
initial set of associations. It will then print out the same
information in various formats.

=back

=cut

######################################################################

sub ClassTest { 
    my @args = @_;
    my $assocList = new CQCC::AssocList();
    my $db = &GetParm("CQCC_DATABASE");
    
    # Test to check for elimination of redundant IDs
    # and extra long numbers and use of commas and/or spaces
    print "Should parse string into " .
            $db . "00000001 " . 
	    $db . "00000155\n";
    $assocList->ProcessBugString(&GetParm("CQCC_ENTITY"),
        $db . "155, 10000000001, 1");
    $assocList->Print();

    print "Should parse string into $db" . "000155, $db" . 
        "000009, $db" . "0000160\n";
    $assocList->ProcessBugString(&GetParm("CQCC_ENTITY"),
        $db . "155 9 160");
    $assocList->Print();   
    print "GetBugList: " . $assocList->GetBugList() . "\n";
    print "GetAssocs: " . join(", ", $assocList->GetAssocs()) . "\n";
    print "GetNumbers: " . join(", ", $assocList->GetNumbers()) . 
       "\n";
    return 0;
}

######################################################################
#::method CreateEditLists

=over 6

=item C<CreateEditLists(desiredList)>

This method compares the current AssocList object ("existing list")
with a second one ("desired list") and produces two new association
lists describing the differences.  The first is the items to be added
to the current list and the second is the items to be deleted from the
current list to produce the "desired list".

($add, $del) = $obj->CreateEditLists($obj2);

=back

=cut

######################################################################

sub CreateEditLists{
    my ($self, $desiredList) = @_;
    my @existingIDs = $self->GetAssocs();
    my @desiredIDs = $desiredList->GetAssocs();
    my (@addIDs,@delIDs, $existing, $desired);
    my $found;
 
    for $existing (@existingIDs) {
        $found = 0;
        for $desired (@desiredIDs) {
            if ($existing eq $desired) {
                $found = 1;
                last;
            }
        }
        if ($found == 0) {
            push @delIDs, $existing;
        }
    }
        
    for $desired (@desiredIDs) {
        $found = 0;
        for $existing (@existingIDs) {
            if ($desired eq $existing) {
                $found = 1;
                last;
            }
        }
        if ($found == 0) { 
            push @addIDs, $desired;
        }
    }
 
    return( new CQCC::AssocList(ASSOCS=>\@addIDs),
            new CQCC::AssocList(ASSOCS=>\@delIDs));
}


######################################################################
#::method GetAssocs

=over 6

=item C<GetAssocs()>

This method returns the internal associations list as a list
consisting of "entity:DBNAMExxxx" entries.

=back

=cut

######################################################################

sub GetAssocs {
    my $self = shift;
    my @assocs;
    if (defined $self->{ASSOCS}) {
        @assocs = @{$self->{ASSOCS}};
    } 
    return @assocs;
}

######################################################################
#::method GetBugList

=over 6

=item C<GetBugList( dbEntityFilter, delimiter )>

This method returns the internal associations list as either a string
or a list depending on the return type expected by the caller.  If a
list is expected it returns a list; otherwise it returns a scalar with
the items delimited by commas similar to what ProcessBugString accepts
for input. All bugs are displayed in full as DBNAME00000xxx and do not
include the entity name.

The dbEntityFilter argument requests that only items matching the
provided database and entity be returned. If undefined then all items
will be returned. 

The optional delimiter argument specifies which delimiter to use to
produce string output. If not specified, a space is used. 

=back

=cut

######################################################################

sub GetBugList {
    my ($self, $dbEntityFilter, $delim) = @_;
    my $str = "";
    my @bugs;
    my ($dbName, $entityName) = split(",", $dbEntityFilter);
    for ($self->GetAssocs()) {
        my ($entity,$bugnum) = split(":",$_);
        if (($entityName && $entity ne $entityName) ||
            ($dbName && 
             substr($bugnum,0,length($bugnum)-8) ne $dbName)) {
            next;
        }
        push @bugs, $bugnum;
    }
    if (wantarray()) {
        return @bugs;
    }
    if (! defined $delim) {
        $delim = " ";
    } 
    return join($delim,@bugs);
}

######################################################################
#::method GetEntityDBList

=over 6

=item C<GetEntityDBList( mergeList )>

This method returns a list of unique "entity:database" combinations
that the association list covers. If a mergeList list is provided then
the return value will be a merge of the provided list and what the
association list has. This is used internally to support handling
lists of mixed entity:database combinations. 

=back

=cut

######################################################################

sub GetEntityDBList {
    my ($self, @entityDBlist) = @_;
    my %entityDBHash;

    #Initialize hash from provided list
    for (@entityDBlist) {
        $entityDBHash{$_} = 1;
    }
    
    my @assocs = $self->GetAssocs();
    for (@assocs) {
        my $entDB = substr($_, 0, length($_)-8);
        $entityDBHash{$entDB} = 1;
    }

    return keys %entityDBHash;
}

######################################################################
#::method GetNumbers

=over 6

=item C<GetNumbers()>

This method returns the internal associations list as a list
consisting of just the numeric bug numbers embedded in the
associations.

=back

=cut

######################################################################

sub GetNumbers {
    my $self = shift;
    my @nums = ();

    foreach (sort $self->GetAssocs()) {
        push @nums, substr($_,-8);
    }
    return @nums;
}

######################################################################
#::method new

=over 6

=item C<new( {ASSOC=E<gt>list} )>

This method creates a new AssocList object, empty by default.  If you
pass in ASSOC=>\@list, it will initialize the association list to the
list that you provide but will not test any internal values.

=back

=cut

######################################################################

sub new { 
    my $proto = shift;
    my $class = ref($proto) || $proto;  # Object or class name

    # Attributes

    my $self = {
       CLASSNAME => "CQCC::AssocList",
       ASSOCS => undef,            # Stored as entity:DBxxxxxx
       @_,                         # Override previous attributes
       @_,                         # Extra time for ccperl
    };

    bless($self, $class);
    return $self;
}

######################################################################
#::method Print

=over 6

=item C<Print()>

This method prints out the current association list. 

=back

=cut

######################################################################

sub Print {
    my $self = shift;
    my $bug;
    my @assocs = $self->GetAssocs();

    &IOPrint("Associations:\n");
    
    for $bug (@assocs) {
        &IOPrint($bug,"\n");
    }
}

######################################################################
#::method ProcessBugString

=over 6

=item C<ProcessBugString(entity, bugstring)>

This method parses user input to determine a list of change request
IDs, assuming that all are in the same database. 

Bugs may be entered with complete names, as 'SAMPL00000008' or
'SAMPL9', or as just numbers '000008' or '9', or any combination
as long as the database name (when used) matches the current 
database. 

=back

=cut

######################################################################

sub ProcessBugString {
    my ($self, $entity, $bugstring) = @_;

    my @bugTokens = &IOParseListString($bugstring);
    my @bugs = ();
    my ($bug,$rbug);
    my $result = 1;
    my @bugnames = ();
    my $db = &GetParm("CQCC_DATABASE");
    my $lenDB = length($db);

    &IODebugPrint("DETAILED", "Starting ProcessBuglist() with @_\n");

    foreach $bug (@bugTokens) {
        $rbug = "";

        # Case 1 - NNNNN - insist that entire pattern is number
        if ($bug =~ /^\d+$/) { 
            $rbug = $bug;
            push @bugs, $rbug;
        }

        # Case 2 - NameNNNNNN - confirm that database and Name match
        # and that remainder of the number is valid numeric only. 
        # Case insensitive and doesn't worry about numbered db names.

        else {
            my $bugDB = substr($bug, 0, $lenDB);
            
            if (uc($db) eq uc($bugDB)) {
               $bug = substr($bug, $lenDB);
               if ($bug =~ /^\d+$/) { 
                  $rbug = $bug;
                  push @bugs, $rbug;
              } else {
                  $result = 0;
              }
            } else {
               $result = 0;
            }
        }
        &IODebugPrint("DETAILED", 
           "ProcessBugString() bug: $bug parsed as $rbug, result: $result\n");
    }

    if ($#bugs && !$db) {
        my @emptyList = ();
        $self->{ASSOCS} = \@emptyList;
        $result = 0;
    }

    if ($result) {
       # commit to saved list
       # Hash list will eliminate multiples and simplify sorting.
       my %bugHash;
       for $bug (@bugs) {
           my $bugLen = length($bug) - 8;
           if ($bugLen < 0) {
               $bugLen = 0;
           }
           my $dbname = sprintf("%s%08d",$db, substr($bug, $bugLen));
           $bugHash{"$entity:$dbname"}=1;
       }
       @bugnames = sort keys %bugHash;
       $self->{ASSOCS} = \@bugnames;
    }

    &IODebugPrint("DETAILED",
        "return from ProcessBugString(): db: $db bugs: \@bugnames " . 
        "result: $result\n");
    
    return $result;
}

######################################################################
#::method SetAssocs

=over 6

=item C<SetAssocs(list)>

This method sets the internal associations list from a list
consisting of "entity:DBNAMExxxx" entries.

=back

=cut

######################################################################

sub SetAssocs {
    my ($self, @assocs) = @_;

    $self->{ASSOCS} = \@assocs;
    return 1;
}

######################################################################
#:: tail
######################################################################

1;
